package graphs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public final class DijkstraAlgorithm {

	private final List<Edge> edges;
	private Set<Vertex> settledNodes;
	private Set<Vertex> unSettledNodes;
	private Map<Vertex, Vertex> predecessors;
	private Map<Vertex, Integer> distance;

	public DijkstraAlgorithm(final Graph graph) {
		new ArrayList<Vertex>(graph.getVertexes());
		this.edges = new ArrayList<Edge>(graph.getEdges());
	}

	public void execute(final Vertex source) {
		settledNodes = new HashSet<Vertex>();
		unSettledNodes = new HashSet<Vertex>();
		distance = new HashMap<Vertex, Integer>();
		predecessors = new HashMap<Vertex, Vertex>();
		distance.put(source, 0);
		unSettledNodes.add(source);
		
		while (unSettledNodes.size() > 0) {
			final Vertex node = getMinimum(unSettledNodes);
			settledNodes.add(node);
			unSettledNodes.remove(node);
			findMinimalDistances(node);
		}
	}

	private void findMinimalDistances(final Vertex node) {
		final List<Vertex> adjacentNodes = getNeighbors(node);
		int adjacent_size = adjacentNodes.size();
		for (int i = 0; i<adjacent_size; i++)
		{
			if (getShortestDistance(adjacentNodes.get(i)) > getShortestDistance(node)
					+ getDistance(node, adjacentNodes.get(i))) {
				distance.put(adjacentNodes.get(i),
						getShortestDistance(node) + getDistance(node, adjacentNodes.get(i)));
				predecessors.put(adjacentNodes.get(i), node);
				unSettledNodes.add(adjacentNodes.get(i));
			}
		}
		
	}

	private int getDistance(final Vertex node, final Vertex target) {
		int edges_size = edges.size();
		for (int i = 0; i<edges_size; i++)
		{
			if (edges.get(i).getSource().equals(node)
					&& edges.get(i).getDestination().equals(target)) {
				return edges.get(i).getWeight();
			}
		}
		throw new RuntimeException("Should not happen");
	}

	private List<Vertex> getNeighbors(final Vertex node) {
		final List<Vertex> neighbors = new ArrayList<Vertex>();
		int edges_size = edges.size();
		for(int i = 0; i<edges_size; i++)
		{
			if (edges.get(i).getSource().equals(node)
					&& !isSettled(edges.get(i).getDestination())) {
				neighbors.add(edges.get(i).getDestination());
			}
		}
		return neighbors;
	}

	private Vertex getMinimum(final Set<Vertex> vertexes) {
		Vertex minimum = null;

		for (Iterator<Vertex> it=vertexes.iterator(); it.hasNext();)  {  
		    Vertex index = (Vertex) it.next();  
		    if (minimum == null) {
				minimum = index;
			} else {
				if (getShortestDistance(index) < getShortestDistance(minimum)) {
					minimum = index;
				}
			}
		}  
		return minimum;
	}

	private boolean isSettled(final Vertex vertex) {
		return settledNodes.contains(vertex);
	}

	private int getShortestDistance(final Vertex destination) {
		final Integer d = distance.get(destination);
		if (d == null) {
			return Integer.MAX_VALUE;
		} else {
			return d;
		}
	}

	public LinkedList<Vertex> getPath(final Vertex target) {
		final LinkedList<Vertex> path = new LinkedList<Vertex>();
		Vertex step = target;
		if (predecessors.get(step) == null) {
			return null;
		}
		path.add(step);
		while (predecessors.get(step) != null) {
			step = predecessors.get(step);
			path.add(step);
		}

		return path;
	}

}